# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Any, Dict

import numpy as np
from openvino.model_api.models.utils import AnomalyResult
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    IPredictionToAnnotationConverter,
)
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map


class AnomalyClassificationToAnnotationConverter(IPredictionToAnnotationConverter):
    """
    Convert AnomalyClassification Predictions ModelAPI to Annotations.

    :param labels: Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]

    def convert_to_annotation(
        self, predictions: np.ndarray, metadata: Dict[str, Any]
    ) -> AnnotationSceneEntity:
        """
        Convert predictions to OTX Annotation Scene using the metadata.

        :param predictions: Raw predictions from the model.
        :param metadata: Variable containing metadata information.
        :return: OTX annotation scene entity object.
        """
        pred_label = predictions >= metadata.get("threshold", 0.5)

        label = self.anomalous_label if pred_label else self.normal_label
        probability = (1 - predictions) if predictions < 0.5 else predictions

        annotations = [
            Annotation(
                Rectangle.generate_full_box(),
                labels=[ScoredLabel(label=label, probability=float(probability))],
            )
        ]
        return AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION, annotations=annotations
        )


class AnomalySegmentationToAnnotationConverter(IPredictionToAnnotationConverter):
    """
    Convert AnomalySegmentation Predictions ModelAPI to Annotations.

    :param labels: Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]
        self.label_map = {0: self.normal_label, 1: self.anomalous_label}

    def convert_to_annotation(
        self, predictions: AnomalyResult, metadata: Dict[str, Any]
    ) -> AnnotationSceneEntity:
        """
        Convert predictions to OTX Annotation Scene using the metadata.

        :param predictions: Raw predictions from the model.
        :param metadata: Variable containing metadata information.
        :return: OTX annotation scene entity object.
        """
        if predictions.pred_mask is None:
            raise ValueError("No prediction mask found in model output")
        if predictions.anomaly_map is None:
            raise ValueError("No anomaly map found in model output")
        annotations = create_annotation_from_segmentation_map(
            predictions.pred_mask, predictions.anomaly_map / 255.0, self.label_map
        )
        if len(annotations) == 0:
            # TODO: add confidence to this label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=self.normal_label, probability=1.0)],
                )
            ]
        return AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION, annotations=annotations
        )


class AnomalyDetectionToAnnotationConverter(IPredictionToAnnotationConverter):
    """
    Convert Anomaly Detection Predictions ModelAPI to Annotations.

    :param labels: Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]
        self.label_map = {0: self.normal_label, 1: self.anomalous_label}

    def convert_to_annotation(
        self, predictions: AnomalyResult, metadata: Dict[str, Any]
    ) -> AnnotationSceneEntity:
        """
        Convert predictions to OTX Annotation Scene using the metadata.

        :param predictions: Raw predictions from the model.
        :param metadata: Variable containing metadata information.
        :return: OTX annotation scene entity object.
        """
        if predictions.pred_mask is None:
            raise ValueError("No prediction mask found in model output")
        if predictions.pred_boxes is None:
            raise ValueError("No bounding boxes found in model output")
        if predictions.pred_score is None:
            raise ValueError("No prediction score found in model output")
        annotations = []
        image_h, image_w = predictions.pred_mask.shape
        for box in predictions.pred_boxes:
            annotations.append(
                Annotation(
                    Rectangle(
                        box[0] / image_w,
                        box[1] / image_h,
                        box[2] / image_w,
                        box[3] / image_h,
                    ),
                    labels=[
                        ScoredLabel(
                            label=self.anomalous_label,
                            probability=predictions.pred_score,
                        )
                    ],
                )
            )
        if len(annotations) == 0:
            # TODO: add confidence to this label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=self.normal_label, probability=1.0)],
                )
            ]
        return AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION, annotations=annotations
        )



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""Module implements the Postprocessor class."""

from typing import List, Tuple

import numpy as np
import otx
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model_template import Domain as OteDomain
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    IPredictionToAnnotationConverter,
    create_converter,
)

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.enums.task_type import TaskType
from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.label_schema import LabelSchema
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.data_models.shapes import Polygon, Rectangle, RotatedRectangle
from geti_sdk.deployment.predictions_postprocessing.utils.detection_utils import (
    detection2array,
)

from .legacy_converters import (
    AnomalyClassificationToAnnotationConverter,
    AnomalyDetectionToAnnotationConverter,
    AnomalySegmentationToAnnotationConverter,
)


class LegacyConverter:
    """
    LegacyConverter class responsible for converting the output of the model to a Prediction object.
    For models generated with Geti v1.8 and below.

    :param labels: Label schema to be used for the conversion.
    :param configuration: Configuration to be used for the conversion.
    :param task: Task object containing the task metadata.
    """

    def __init__(
        self, label_schema: LabelSchema, configuration, domain: Domain
    ) -> None:
        self.domain = domain
        self.task_type = TaskType[self.domain.name]
        self.label_schema = LabelSchemaEntity(
            label_groups=[
                LabelGroup(
                    name=group.name,
                    labels=[label.to_ote(self.task_type) for label in group.labels],
                    group_type=LabelGroupType[group.group_type.name],
                    id=group.id,
                )
                for group in label_schema.get_groups(include_empty=True)
            ]
        )

        # Create OTX converter
        converter_args = {"labels": self.label_schema}
        if otx.__version__ > "1.2.0":
            if "use_ellipse_shapes" not in configuration.keys():
                configuration.update({"use_ellipse_shapes": False})
            converter_args["configuration"] = configuration

        self.converter: IPredictionToAnnotationConverter = create_converter(
            converter_type=OteDomain[self.domain.name], **converter_args
        )

    def convert_to_prediction(
        self, postprocessing_results: List, image_shape: Tuple[int], **kwargs
    ) -> Prediction:
        """
        Convert the postprocessing results to a Prediction object.
        """
        # Handle empty annotations
        if isinstance(postprocessing_results, (np.ndarray, list)):
            try:
                n_outputs = len(postprocessing_results)
            except TypeError:
                n_outputs = 1
        else:
            # Handle the new modelAPI output formats for detection and instance
            # segmentation models
            if (
                hasattr(postprocessing_results, "objects")
                and self.domain == Domain.DETECTION
            ):
                n_outputs = len(postprocessing_results.objects)
                postprocessing_results = detection2array(postprocessing_results.objects)
            elif hasattr(
                postprocessing_results, "segmentedObjects"
            ) and self.domain in [
                Domain.INSTANCE_SEGMENTATION,
                Domain.ROTATED_DETECTION,
            ]:
                n_outputs = len(postprocessing_results.segmentedObjects)
                postprocessing_results = postprocessing_results.segmentedObjects
            elif isinstance(postprocessing_results, tuple):
                try:
                    n_outputs = len(postprocessing_results)
                except TypeError:
                    n_outputs = 1
            else:
                raise ValueError(
                    f"Unknown postprocessing output of type "
                    f"`{type(postprocessing_results)}` for task `{self.task.title}`."
                )

        # Proceed with postprocessing
        width: int = image_shape[1]
        height: int = image_shape[0]

        if n_outputs != 0:
            try:
                annotation_scene_entity = self.converter.convert_to_annotation(
                    predictions=postprocessing_results,
                    metadata={"original_shape": image_shape},
                )
            except AttributeError as e:
                # Add backwards compatibility for anomaly models created in Geti v1.8 and below
                if self.domain == Domain.ANOMALY_CLASSIFICATION:
                    legacy_converter = AnomalyClassificationToAnnotationConverter(
                        label_schema=self.label_schema
                    )
                elif self.domain == Domain.ANOMALY_DETECTION:
                    legacy_converter = AnomalyDetectionToAnnotationConverter(
                        label_schema=self.label_schema
                    )
                elif self.domain == Domain.ANOMALY_SEGMENTATION:
                    legacy_converter = AnomalySegmentationToAnnotationConverter(
                        label_schema=self.label_schema
                    )
                else:
                    raise e

                annotation_scene_entity = legacy_converter.convert_to_annotation(
                    predictions=postprocessing_results,
                    metadata={"original_shape": image_shape},
                )
                self.converter = legacy_converter

            prediction = Prediction.from_ote(
                annotation_scene_entity, image_width=width, image_height=height
            )
        else:
            prediction = Prediction(annotations=[])

        # Empty label is not generated by OTE correctly, append it here if there are
        # no other predictions
        if len(prediction.annotations) == 0:
            empty_label = next(
                (
                    label
                    for label in self.label_schema.get_labels(include_empty=True)
                    if label.is_empty
                ),
                None,
            )
            if empty_label is not None:
                empty_label_sdk = Label(
                    name=empty_label.name,
                    color=(
                        empty_label.color
                        if isinstance(empty_label.color, str)
                        else empty_label.color.hex_str
                    ),
                    id=empty_label.id_,
                    is_empty=True,
                    group="",
                )
                prediction.append(
                    Annotation(
                        shape=Rectangle(x=0, y=0, width=width, height=height),
                        labels=[ScoredLabel.from_label(empty_label_sdk, probability=1)],
                    )
                )

        # Rotated detection models produce Polygons, convert them here to
        # RotatedRectangles
        if self.domain == Domain.ROTATED_DETECTION:
            for annotation in prediction.annotations:
                if isinstance(annotation.shape, Polygon):
                    annotation.shape = RotatedRectangle.from_polygon(annotation.shape)

        return prediction



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time
import warnings
from json import JSONDecodeError
from typing import Any, Dict, Optional, Union

import requests
import simplejson
import urllib3
from requests import Response
from requests.exceptions import RequestException
from requests.structures import CaseInsensitiveDict
from urllib3.exceptions import InsecureRequestWarning

from geti_sdk.platform_versions import GETI_18_VERSION, GETI_114_VERSION, GetiVersion

from .exception import GetiRequestException
from .server_config import LEGACY_API_VERSION, ServerCredentialConfig, ServerTokenConfig

CSRF_COOKIE_NAME = "_oauth2_proxy_csrf"
PROXY_COOKIE_NAME = "_oauth2_proxy"
GETI_COOKIE_NAME = "geti-cookie"

# INITIAL_HEADERS = {"Connection": "keep-alive", "Upgrade-Insecure-Requests": "1"}
INITIAL_HEADERS = {"Upgrade-Insecure-Requests": "1"}
SUCCESS_STATUS_CODES = [200, 201, 202]

AUTHENTICATION_DEX_OLD = "dex-old"
AUTHENTICATION_DEX_NEW = "dex-new"
AUTHENTICATION_CIDAAS = "cidaas"


class GetiSession(requests.Session):
    """
    Wrapper for requests.session that sets the correct headers and cookies, and
    handles authentication and authorization.

    :param server_config: Server configuration holding the hostname (or ip address) of
        the Intel® Geti™ server, as well as the details required for authentication
        (either username and password or personal access token)
    """

    def __init__(
        self,
        server_config: Union[ServerTokenConfig, ServerCredentialConfig],
    ):
        super().__init__()
        self.headers.update(INITIAL_HEADERS)
        self.allow_redirects = False
        self.token = None
        self._cookies: Dict[str, Optional[str]] = {
            CSRF_COOKIE_NAME: None,
            PROXY_COOKIE_NAME: None,
        }

        # Configure proxies
        if server_config.proxies is None:
            self._proxies: Dict[str, str] = {}
        else:
            self._proxies = {"proxies": server_config.proxies}

        # Configure certificate verification
        if not server_config.has_valid_certificate:
            warnings.warn(
                "You have disabled TLS certificate validation, HTTPS requests made to "
                "the Intel® Geti™ server may be compromised. For optimal security, "
                "please enable certificate validation.",
                InsecureRequestWarning,
            )
            urllib3.disable_warnings(InsecureRequestWarning)
        self.verify = server_config.has_valid_certificate

        self.config = server_config
        self.logged_in = False
        self._auth_service: Optional[str] = None

        # Determine authentication method
        if isinstance(server_config, ServerCredentialConfig):
            self.authenticate(verbose=True)
            self.use_token = False
        else:
            self.use_token = True
            if self.authentication_service == AUTHENTICATION_DEX_OLD:
                # The old token mechanism
                access_token = self._acquire_access_token()
                self.headers.update({"Authorization": f"Bearer {access_token}"})
            else:
                # New mechanism, from Geti v1.14
                self.headers.update({"x-api-key": f"{server_config.token}"})
            self.headers.pop("Connection")

        # Get server version
        self._product_info = self._get_product_info_and_set_api_version()
        self._organization_id: Optional[str] = self._get_organization_id()

    @property
    def version(self) -> GetiVersion:
        """
        Return the version of the Intel® Geti™ platform that is running on the server.

        :return: Version object holding the Intel® Geti™ version number
        """
        if "build-version" in self._product_info.keys():
            version_string = self._product_info.get("build-version")
        else:
            version_string = self._product_info.get("product-version")
        return GetiVersion(version_string)

    @property
    def organization_id(self) -> str:
        """
        Return the organization id for the user who is associated with the GetiSession
        """
        if self._organization_id is None:
            self._organization_id = self._get_organization_id()
        return self._organization_id

    def _acquire_access_token(self) -> str:
        """
        Request an access token from the server, in exchange for the
        PersonalAccessToken.
        """
        try:
            response = self.get_rest_response(
                url="service_accounts/access_token",
                method="POST",
                data={"service_id": self.config.token},
                contenttype="json",
                allow_reauthentication=False,
                include_organization_id=False,
            )
        except GetiRequestException as error:
            if error.status_code == 401:
                raise ValueError(
                    "Token authorization failed. Please make sure to provide a valid "
                    "Personal Access Token."
                )
            raise
        logging.info(f"Personal access token validated on host {self.config.host}")
        return response.get("access_token")

    def _follow_login_redirects(self, response: Response, use_legacy_dex: bool) -> str:
        """
        Recursively follow redirects in the initial login request. Updates the
        session._cookies with the cookie and the login uri.

        :param response: REST response to follow redirects for
        :param use_legacy_dex: True to use the old authentication method, used
                    for DEX in Geti v1.8 and below. False to use the new method, which
                    is valid for more recent Geti versions
        :return: url to the redirected location
        """
        if use_legacy_dex:
            if response.status_code in [302, 303]:
                redirect_url = response.next.url
                redirected = self.get(
                    redirect_url, allow_redirects=False, **self._proxies
                )
                proxy_csrf = redirected.cookies.get(CSRF_COOKIE_NAME, None)
                if proxy_csrf is None:
                    proxy_csrf = response.cookies.get(CSRF_COOKIE_NAME, None)
                if proxy_csrf is not None:
                    self._cookies[CSRF_COOKIE_NAME] = proxy_csrf
                return self._follow_login_redirects(redirected, use_legacy_dex)
            else:
                return response.url
        else:
            if response.status_code in [302, 303]:
                return response.next.url

    def _get_initial_login_url(self, use_legacy_dex: bool) -> str:
        """
        Retrieve the initial login url by making a request to the login page, and
        following the redirects.

        :param use_legacy_dex: True to use the old authentication method, used
                    for DEX in Geti v1.8 and below. False to use the new method, which
                    is valid for more recent Geti versions
        :return: string containing the URL to the login page
        """
        if use_legacy_dex:
            url = self.config.host
        else:
            params = (
                "client_id=web_ui"
                "&redirect_uri=/callback"
                "&code_challenge_method=S256"
                "&response_type=code"
                "&response_mode=query"
                "&scope=openid+profile+groups+email+offline_access"
            )
            url = f"{self.config.host}/dex/auth/regular_users?{params}"
        response = self.get(url, allow_redirects=False, **self._proxies)
        login_page_url = self._follow_login_redirects(response, use_legacy_dex)
        return login_page_url

    def authenticate(self, verbose: bool = True):
        """
        Get a new authentication cookie from the server.

        :param verbose: True to print progress output, False to suppress output
        """
        if self.logged_in:
            logging.info("Already logged in, authentication is skipped")
            return
        if self.authentication_service == AUTHENTICATION_DEX_OLD:
            use_legacy_dex = True
        elif self.authentication_service == AUTHENTICATION_DEX_NEW:
            use_legacy_dex = False
        else:
            raise ValueError(
                f"Unable to start the authentication process: The Geti server "
                f"requires an authentication service that this version of the "
                f"geti-sdk does not support: {self.authentication_service}"
            )
        self.cookies.clear()
        try:
            login_path = self._get_initial_login_url(use_legacy_dex)
        except requests.exceptions.SSLError as error:
            raise requests.exceptions.SSLError(
                f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                f"the server address can be resolved but the SSL certificate could not "
                f"be verified. \n Full error description: {error.args[-1]}"
            )
        except requests.exceptions.ConnectionError as error:
            if "dummy" in self.config.password or "dummy" in self.config.username:
                raise ValueError(
                    "Connection to the Intel® Geti™ server failed, please make sure to "
                    "update the user login information for the Intel® Geti™ instance."
                ) from error
            raise ValueError(
                f"Connection to the Intel® Geti™ server at host '{self.config.host}' "
                f"failed, please provide a valid cluster hostname or ip address as"
                f" well as valid login details."
            ) from error
        self.headers.clear()
        self.headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        if verbose:
            logging.info(f"Authenticating on host {self.config.host}...")
        if use_legacy_dex:
            cookies = {CSRF_COOKIE_NAME: self._cookies.get(CSRF_COOKIE_NAME, None)}
            headers = {"Cookie": self._cookies.get(CSRF_COOKIE_NAME, None)}
        else:
            cookies = {}
            headers = self.headers
        response = self.post(
            url=login_path,
            data={"login": self.config.username, "password": self.config.password},
            cookies=cookies,
            headers=headers,
            allow_redirects=True,
            **self._proxies,
        )
        if use_legacy_dex:
            self._handle_legacy_dex_response(response)
        else:
            self._handle_dex_response(response)
        if verbose:
            logging.info("Authentication successful. Cookie received.")
        self.logged_in = True

    def get_rest_response(
        self,
        url: str,
        method: str,
        contenttype: str = "json",
        data=None,
        allow_reauthentication: bool = True,
        include_organization_id: bool = True,
        allow_text_response: bool = False,
    ) -> Union[Response, dict, list]:
        """
        Return the REST response from a request to `url` with `method`.

        :param url: the REST url without the hostname and api pattern
        :param method: 'GET', 'POST', 'PUT', 'DELETE'
        :param contenttype: currently either 'json', 'jpeg', 'multipart', 'zip', or '',
            defaults to "json"
        :param data: the data to send in a post request, as json
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        :param include_organization_id: True to include the organization ID in the base
            URL. Can be set to False for accessing certain internal endpoints that do
            not require an organization ID, but do require error handling.
        :param allow_text_response: False to trigger error handling when the server
            returns a response with text/html content. This can happen in some cases
            when authentication has expired. However, some endpoints are designed to
            return text responses, for those endpoints this parameter should be set to
            True
        """
        if url.startswith(self.config.api_pattern):
            url = url[len(self.config.api_pattern) :]

        self._update_headers_for_content_type(content_type=contenttype)

        if not include_organization_id or url.startswith(
            f"organizations/{self.organization_id}/"
        ):
            requesturl = f"{self.config.base_url}{url}"
        else:
            requesturl = f"{self.base_url}{url}"

        if method == "POST" or method == "PUT":
            if contenttype == "json":
                kw_data_arg = {"json": data}
            elif contenttype == "multipart":
                kw_data_arg = {"files": data}
            elif contenttype == "jpeg" or contenttype == "zip":
                kw_data_arg = {"data": data}
            else:
                raise ValueError(
                    f"Making a POST request with content of type {contenttype} is "
                    f"currently not supported through the Geti SDK."
                )
        else:
            kw_data_arg = {}

        request_params = {
            "method": method,
            "url": requesturl,
            **kw_data_arg,
            "stream": True,
        }

        if not self.use_token:
            request_params.update({"cookies": self._cookies})

        csrf_header_methods = ["POST", "PUT", "DELETE", "PATCH"]
        if method in csrf_header_methods:
            self.headers.update({"x-geti-csrf-protection": "1"})
        else:
            self.headers.pop("x-geti-csrf-protection", "")

        try:
            response = self.request(**request_params, **self._proxies)
        except requests.exceptions.SSLError as error:
            raise requests.exceptions.SSLError(
                f"Connection to Intel® Geti™ server at '{self.config.host}' failed, "
                f"the server address can be resolved but the SSL certificate could not "
                f"be verified. \n Full error description: {error.args[-1]}"
            )
        except ConnectionError as conn_error:
            if conn_error.args[0] == "Connection aborted.":
                # We fake a response and try to establish a
                # new connection by re-authenticating
                response = Response()
                response.status_code = 401
                response.raw = conn_error.args[-1]
            else:
                raise conn_error

        response_content_type = response.headers.get("Content-Type", [])
        if (
            response.status_code not in SUCCESS_STATUS_CODES
            or "text/html" in response_content_type
        ):
            if not ("text/html" in response_content_type and allow_text_response):
                response = self._handle_error_response(
                    response=response,
                    request_params=request_params,
                    request_data=kw_data_arg,
                    allow_reauthentication=allow_reauthentication,
                    content_type=contenttype,
                )

        if response.headers.get("Content-Type", "").startswith("application/json"):
            result = response.json()
        else:
            result = response

        return result

    def logout(self, verbose: bool = True) -> None:
        """
        Log out of the server and end the session. All HTTPAdapters are closed and
        cookies and headers are cleared.

        :param verbose: False to suppress output messages
        """
        requires_closing = True
        requires_sign_out = True

        if not self.logged_in or self.use_token:
            requires_closing = False
            requires_sign_out = False

        if requires_sign_out:
            sign_out_url = (
                self.config.base_url[: -len(self.config.api_pattern)]
                + "/oauth2/sign_out"
            )
            try:
                response = self.request(url=sign_out_url, method="GET", **self._proxies)

                if response.status_code in SUCCESS_STATUS_CODES:
                    if verbose:
                        logging.info("Logout successful.")
                else:
                    raise GetiRequestException(
                        method="GET",
                        url=sign_out_url,
                        status_code=response.status_code,
                        request_data={},
                    )
            except (RequestException, AttributeError):
                if verbose:
                    logging.info(
                        f"The {self.__class__.__name__} is closed successfully, but "
                        f"the Geti instance was not able to logout from the server."
                    )
            except TypeError:
                # The session is already closed
                requires_closing = False

        self._cookies = {}
        self.cookies.clear()
        self.headers = CaseInsensitiveDict(INITIAL_HEADERS)
        self.logged_in = False
        if requires_closing:
            try:
                super().close()
            except TypeError as error:
                # Sometimes a TypeError is raised if logout is called during garbage
                # collection, however we can safely ignore this since the session will
                # be deleted in that case anyway.
                logging.debug(f"{error} encountered during GetiSession closure.")

    def _get_product_info_and_set_api_version(self) -> Dict[str, str]:
        """
        Return the product info as retrieved from the Intel® Geti™ server.

        This method will also attempt to set the API version correctly, based on the
        retrieved product info.

        :return: Dictionary containing the product info.
        """
        try:
            product_info = self.get_rest_response(
                "product_info", "GET", include_organization_id=False
            )
        except GetiRequestException:
            self.config.api_version = LEGACY_API_VERSION
            product_info = self.get_rest_response(
                "product_info", "GET", include_organization_id=False
            )
        return product_info

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Log out of the Intel® Geti™ server. This method is called when exiting the
        runtime context related to the session, in case the session is used as a context
        manager.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered while exiting GetiSession context.")
        super().__exit__(exc_type, exc_value, traceback)

    def __del__(self):
        """
        Log out of the Intel® Geti™ server. This method is called when the session is
        deleted from memory.
        """
        if self.logged_in:
            try:
                self.logout(verbose=False)
            except Exception as exc:
                logging.debug(f"{exc} encountered during GetiSession deletion.")

    def _handle_error_response(
        self,
        response: Response,
        request_params: Dict[str, Any],
        request_data: Dict[str, Any],
        allow_reauthentication: bool = True,
        content_type: str = "json",
    ) -> Response:
        """
        Handle error responses from the server.

        :param response: The Response object received from the server
        :param request_params: Dictionary containing the original parameters of the
            request
        :param allow_reauthentication: True to handle authentication errors
            by attempting to re-authenticate. If set to False, such errors
            will be raised instead.
        :param content_type: The content type of the original request
        :raises: GetiRequestException in case the error cannot be handled
        :return: Response object resulting from the request
        """
        retry_request = False

        if response.status_code in [200, 401] and allow_reauthentication:
            # Authentication has likely expired, re-authenticate
            logging.info("Authentication may have expired, re-authenticating...")
            self.logged_in = False
            if not self.use_token:
                self.authenticate(verbose=False)
                logging.info("Authentication complete.")

            else:
                if self.authentication_service == AUTHENTICATION_DEX_OLD:
                    access_token = self._acquire_access_token()
                    logging.info("New bearer token obtained.")
                    self.headers.update({"Authorization": f"Bearer {access_token}"})
                else:
                    raise ValueError(
                        "Authentication via your personal access token has failed, "
                        "most likely the token has expired. Please verify that you "
                        "have provided a valid token."
                    )

            retry_request = True

        elif response.status_code == 503:
            # In case of Service Unavailable, wait some time and try again. If it
            # still doesn't work, raise exception
            time.sleep(1)
            retry_request = True

        # We make one attempt to do the request again. If it fails again, a
        # GetiRequestException will be raised holding further details of the
        # reason for failure.
        if retry_request:
            self._update_headers_for_content_type(content_type=content_type)
            # Reset any file buffers that were included in the request data, so that we
            # can attempt to upload them again.
            if content_type == "multipart":
                for file_name, file_buffer in request_params["files"].items():
                    file_buffer.seek(0, 0)

            response = self.request(**request_params, **self._proxies)

            if response.status_code in SUCCESS_STATUS_CODES:
                return response

        try:
            response_data = response.json()
        except (JSONDecodeError, simplejson.errors.JSONDecodeError):
            response_data = None

        raise GetiRequestException(
            method=request_params["method"],
            url=request_params["url"],
            status_code=response.status_code,
            request_data=request_data,
            response_data=response_data,
        )

    def _update_headers_for_content_type(self, content_type: str) -> None:
        """
        Update the session headers to contain the correct content type

        :param content_type: content type for the request
        """
        if content_type == "json":
            self.headers.update({"Content-Type": "application/json"})
        elif content_type == "jpeg":
            self.headers.update({"Content-Type": "image/jpeg"})
        elif content_type == "multipart":
            self.headers.pop("Content-Type", None)
        elif content_type == "":
            self.headers.pop("Content-Type", None)
        elif content_type == "zip":
            self.headers.update({"Content-Type": "application/zip"})

    @property
    def base_url(self) -> str:
        """
        Return the base URL to the Intel Geti server. If the server is running
        Geti v1.9 or later, the organization ID will be included in the URL
        """
        if self.version <= GETI_18_VERSION:
            return self.config.base_url
        else:
            return f"{self.config.base_url}organizations/{self.organization_id}/"

    @property
    def authentication_service(self) -> str:
        """
        Return the type of authentication service used by the server to which
        the GetiSession is tied.
        """
        if self._auth_service is None:
            deployment_config_response = self.request(
                url=f"{self.config.host}/deployment-config.json",
                method="GET",
                **self._proxies,
            )
            try:
                authentication_info = deployment_config_response.json().get(
                    "auth", {"type": "dex"}
                )
            except requests.exceptions.JSONDecodeError:
                self._auth_service = AUTHENTICATION_DEX_OLD
                return self._auth_service
            if authentication_info.get("type") == "dex":
                self._auth_service = AUTHENTICATION_DEX_NEW
            else:
                self._auth_service = AUTHENTICATION_CIDAAS
        return self._auth_service

    def _get_organization_id(self) -> str:
        """
        Return the organization ID associated with the user and host information configured
        in this Session
        """
        default_org_id = "000000000000000000000001"
        if self.version < GETI_114_VERSION:
            org_id = default_org_id
        else:
            if not self.use_token:
                result = self.get_rest_response(
                    url="profile",
                    method="GET",
                    include_organization_id=False,
                )
            else:
                result = self.get_rest_response(
                    url="personal_access_tokens/organization",
                    method="GET",
                    include_organization_id=False,
                )
            org_id = result.get("organizationId", None)
        if org_id is None:
            raise ValueError(
                f"Unable to retrieve organization ID from the Intel Geti server. "
                f"Server responded with: `{result}`"
            )
        return org_id

    def _handle_legacy_dex_response(self, response: Response) -> None:
        """
        Process the login response for the old-style DEX authorization mechanism

        :param response: Response object to handle
        """
        try:
            previous_response = response.history[-1]
        except IndexError:
            raise ValueError(
                "The cluster responded to the request, but authentication failed. "
                "Please verify that you have provided correct credentials."
            )
        proxy_cookie = previous_response.cookies.get(PROXY_COOKIE_NAME)
        if proxy_cookie is not None:
            cookie = {PROXY_COOKIE_NAME: proxy_cookie}
            self._cookies.update(cookie)
        else:
            logging.warning(
                f"Authentication appears to have failed! No valid oauth cookie "
                f"obtained. Invalid response received from server. Status code: "
                f"{response.status_code}"
            )

    def _handle_dex_response(self, response: Response) -> None:
        """
        Process the login response from the DEX authorization mechanism

        Extract the code and exchange it for a token, which must then
        be used as authentication cookie

        :param response: DEX response object to handle
        """
        if response.status_code == 200:
            code = response.url.split("code=")[1].split("&")[0]
        elif response.status_code == 401:
            raise ValueError(
                "The cluster responded to the request, but authentication failed. "
                "Please verify that you have provided correct credentials."
            )
        else:
            raise GetiRequestException(
                method=response.request.method,
                url=response.url,
                status_code=response.status_code,
                request_data={},
                response_data={
                    "message": response.content,
                    "error_code": response.reason,
                },
            )
        # state = login_path.split("state=")[1].split("&")[0]
        data = {
            "grant_type": "authorization_code",
            "redirect_uri": "/callback",
            "code": code,
            "client_id": "web_ui",
        }
        login_response = self.post(
            url=f"{self.config.host}/dex/token",
            data=data,
            allow_redirects=True,
            **self._proxies,
        )
        token = login_response.json().get("access_token", None)
        self._cookies.update({GETI_COOKIE_NAME: token})
